
# THIS FILE IS GENERATED FROM HIFIVE SETUP.PY
short_version = '1.5.7'
version = '1.5.7'
full_version = '1.5.7'
git_revision = '317d2dc98bdfb13ed4d528f718b9492f7c12aabd'
release = True

if not release:
    version = full_version
